use crate::{
    schema::{AssocData, Datum},
    DBResult,
};

impl Datum for String {
    fn sql_type() -> &'static str {
        "text"
    }

    fn bind_to<'a>(&self, stmt: &mut sqlite::Statement<'a>, index: usize) {
        stmt.bind((index, self.as_str()))
            .expect("couldn't bind string");
    }

    fn build_from<'a>(
        _: AssocData,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read(index)?, index + 1))
    }
}

impl Datum for usize {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, stmt: &mut sqlite::Statement<'a>, index: usize) {
        stmt.bind((index, *self as i64))
            .expect("couldn't bind usize as i64");
    }

    fn build_from<'a>(
        _: AssocData,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read::<i64, _>(index)? as usize, index + 1))
    }
}

impl Datum for isize {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        todo!()
    }

    fn build_from<'a>(
        _: AssocData,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read::<i64, _>(index)? as isize, index + 1))
    }
}

impl Datum for u64 {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        todo!()
    }

    fn build_from<'a>(
        _: AssocData,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read::<i64, _>(index)? as u64, index + 1))
    }
}

impl Datum for i64 {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        todo!()
    }

    fn build_from<'a>(
        _: AssocData,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read(index)?, index + 1))
    }
}

impl<T: Datum> Datum for Option<T> {
    fn sql_type() -> &'static str {
        T::sql_type()
    }

    fn bind_to<'a>(&self, _stmt: &mut sqlite::Statement<'a>, _index: usize) {
        todo!()
    }

    fn build_from<'a>(
        _: AssocData,
        _stmt: &mut sqlite::Statement<'a>,
        _index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        // Ok((stmt.read::<i64, _>(index)? as u64, index+1))
        todo!()
    }
}

impl Datum for bool {
    fn sql_type() -> &'static str {
        "int"
    }

    fn bind_to<'a>(&self, stmt: &mut sqlite::Statement<'a>, index: usize) {
        stmt.bind((index, if *self { 1 } else { 0 }))
            .expect("couldn't bind bool");
    }

    fn build_from<'a>(
        _: AssocData,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read::<i64, _>(index)? == 1, index + 1))
    }
}

impl Datum for Vec<u8> {
    fn sql_type() -> &'static str {
        "blob"
    }

    fn bind_to<'a>(&self, stmt: &mut sqlite::Statement<'a>, index: usize) {
        stmt.bind((index, self.as_slice()))
            .expect("couldn't bind Vec<u8>");
    }

    fn build_from<'a>(
        _: AssocData,
        stmt: &mut sqlite::Statement<'a>,
        index: usize,
    ) -> DBResult<(Self, usize)>
    where
        Self: Sized,
    {
        Ok((stmt.read::<Vec<u8>, _>(index)?, index + 1))
    }
}
